package ui;

import java.awt.event.*;
import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Panel;
import java.awt.Choice;
import java.awt.Image;
import java.awt.Color;
import java.awt.Point;
import java.awt.Font;
import java.util.*;




import world.*;


public class MainPanel extends Panel implements ButtonsInterface, ActionListener{

	World world = World.world;
	
	MainPanel(){
		
		MouseDispatcher mouseDispatcher = new MouseDispatcher();
		addMouseListener(mouseDispatcher);
		addMouseMotionListener(mouseDispatcher);
		addKeyListener(mouseDispatcher);
		
		setPreferredSize(new Dimension(640, 480));
		
		(new javax.swing.Timer(1000 / 15, this)).start();
		
	}
	
	public void actionPerformed(ActionEvent event) {
		Simulator.tick(); 
		repaint();
	}
	
	
	
	//public Choice choice = new Choice();
	
	public void initializeButtons(Buttons buttons){
		buttons.addButton("start timer");
		buttons.addButton("forward");
		buttons.addButton("left");
		buttons.addButton("right");
		buttons.addButton("stop");
		buttons.addButton("goto mark");
		buttons.addButton("deliver");
	}
	public void buttonPressed(String label){
		if (label.equals("start timer")){
			start_time = System.currentTimeMillis();
			current_time = start_time;
			timer_is_running = true;
		}
		else if (label.equals("stop"))
			world.get_robot().stop();
		else if (label.equals("forward"))
			world.get_robot().forward();
		else if (label.equals("left"))
			world.get_robot().spin_left();
		else if (label.equals("right"))
			world.get_robot().spin_right();
		else if (label.equals("goto mark"))
			control.MoveToController.moveto(world.get_robot(), world.get_mark());
		else if (label.equals("deliver"))
			control.DeliveryController.deliver(world.get_robot(), world.get_item(), world.get_mark());
		
		repaint();
	}	
	

	
	
	
	
	
	
	
	
	Image buffered_image = null;
	public void update(Graphics g) {
		paint(g);
	}
	Font font = new Font("Helvetica", Font.PLAIN, 24);
	public void paint(Graphics g) {

			//double buffering
			if (buffered_image == null)
				buffered_image = createImage(getSize().width, getSize().height);
			Graphics bg = buffered_image.getGraphics();

			// clear
			bg.setColor(new Color(158,255,158));
			bg.fillRect(0,0, getSize().width, getSize().height);

			bg.setFont(font);
			
			// paint code!
			paint_world(bg);
			paint_stroke(bg);
			paint_time(bg);
			
			
			
			
			//double buffering
			bg.dispose();
      
			g.drawImage(buffered_image,0,0,this);
		
	}
	
   
	
	
	
	
	//
	// Do not change
	//
	public void paint_world(Graphics g){
		for(Mark mark : world.marks){
			mark.paint(g);
		}
		for(Item item : world.items){
			item.paint(g);
		}
		for(Robot robot : world.robots){
			robot.paint(g);
		}
		
	}		
	
	
	
	
	//
	//	timer
	//
	long start_time = System.currentTimeMillis();
	long current_time = start_time;
	boolean timer_is_running;
	void paint_time(Graphics g){
		g.setColor(Color.red);
		if (check_arrival()){
			g.drawString("ARRIVED!", 10, 48);
			timer_is_running = false;
		}
		if (timer_is_running)
			current_time = System.currentTimeMillis();
		g.drawString("Elapsed Time: "+(current_time-start_time)/1000, 10, 24);
	}
	public boolean check_arrival(){
		double d = Vector2D.distance(world.get_item().position, world.get_mark().position);
		return (d<15);
	}
	
	
	
	
	
	//
	//	utility (directly dragging entities.) should be turned off in the user test.
	//
	static boolean drag_enabled = true;
	static public Entity picked_entity = null;
	Vector2D diff;
	boolean start_dragging(Point p){
		if (!drag_enabled)
			return false;
		
		picked_entity = world.pick_entity(new Vector2D(p));
		if (picked_entity == null)
			return false;
		
		diff = new Vector2D(new Vector2D(p), picked_entity.position);
		return true;
	}
	void dragging(Point p){
		Vector2D v = new Vector2D(p);
		picked_entity.position.set(Vector2D.add(v, diff));
	}
	
	
	
	
	
	//
	// sample code for user interaction
	//
	public void paint_stroke(Graphics g){
		if (stroke == null)
			return;
		for(int i=0; i<stroke.size()-1; i++){
			Point p = (Point) stroke.get(i);
			Point q = (Point) stroke.get(i+1);
			g.drawLine(p.x, p.y, q.x, q.y);
		}
	}
	ArrayList<Point> stroke = null;
	public void start_stroke(Point p){
		stroke = new ArrayList<Point> ();
		extend_stroke(p);
	}
	public void extend_stroke(Point p){
		stroke.add(p);
		repaint();
	}
	public void finish_stroke(Point p){
		stroke = null;
	}
	
	
	
	
	
	
	
	

	
	
	
	
	
	
	
   
	
	public int operation_status;
	public static final int NONE = 0;
	public static final int DRAWING = 1;
	public static final int DRAGGING = 2;
   
	//   
	// central event dispatcher
	//
	public class MouseDispatcher 
		extends MouseAdapter implements MouseMotionListener, KeyListener{

  		public void mouseMoved(MouseEvent e) {}
		public void mouseClicked(MouseEvent e){}
  		public void mousePressed(MouseEvent e) {
			Point p = e.getPoint();
			
			if (start_dragging(p)){
				operation_status = DRAGGING;
			}
			else {
				operation_status = DRAWING;
				start_stroke(p);
			}
		}
  		public void mouseDragged(MouseEvent e) {
			Point p = e.getPoint();
			
			switch(operation_status){
			case DRAGGING:
				dragging(p);
				break;
			case DRAWING:
				extend_stroke(p);
				break;
			}
			
   		}
		public void mouseReleased(MouseEvent e) {
			Point p = e.getPoint();
			
			switch(operation_status){
			case DRAWING:
				finish_stroke(p);
				break;
			}
			
			picked_entity = null;
		}

		public void keyTyped(KeyEvent e){}
		public void keyPressed(KeyEvent e){
			System.out.println("key pressed: "+e.getKeyCode());
		}
		public void keyReleased(KeyEvent e){}
    }
	
	
	static public void main(String[] args) {
	
		f = new Frame("Main");
		f.addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent e) {
				System.exit(0);
			}
		});
		f.setMenuBar(new MyMenuBar());
		
		panel = new Panel();
		panel.setLayout(new BorderLayout());
		mainPanel = new MainPanel();
		panel.add("Center", mainPanel);
		panel.add("South", new Buttons(mainPanel));
		f.add("Center", panel);
		f.setVisible(true);
		f.pack();
	}

	static Frame f;
	static Panel panel;
	static MainPanel mainPanel;
	
}







