package world;

import java.awt.event.*;
import java.util.*;



import VisualNumerics.math.DoubleLU;
import VisualNumerics.math.DoubleMatrix;
import VisualNumerics.math.DoubleQR;


//
// do not modify!
//
public class Simulator {
	static World world;
	static void init(World _world){
		world = _world;
		
	}
		
		
	static public void tick(){
		for(Robot robot : world.robots){
			tick_robot(robot);
		}
		for(int i=0; i<100; i++)
			resolve_collisions_static_forces();
		
		notify_simulation_cycle();
	}

	

	
	//
	// multi-thread synchronization
	//
	static public Simulator simulator = new Simulator();
	public static void notify_simulation_cycle() {
		simulator.notify_simulation_cycle_main();
	}
	public static void wait_next_simulation_cycle() {
		simulator.wait_next_simulation_cycle_main();
	}
	synchronized protected void notify_simulation_cycle_main() {
		notifyAll();
	}
	synchronized protected void wait_next_simulation_cycle_main() {
		notifyAll();
		try {
			wait();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	
	
	
	
	
	
	static private double ANGLE_NOISE = 5;
	static public void tick_robot(Robot robot){
		switch (robot.current_motion) {
		case Robot.FORWARD:
		{
			double velocity = 5 + (Math.random() - 0.5) * 2;
			robot.position.x = robot.position.x + velocity
					* robot.orientation.x;
			robot.position.y = robot.position.y + velocity
					* robot.orientation.y;

			double angle = Vector2D.get_angle_360(Vector2D.X_AXIS,
					robot.orientation);
			angle += (Math.random() - 0.5) * 2*ANGLE_NOISE;
			robot.orientation.x = Math.cos(angle * Math.PI / 180);
			robot.orientation.y = Math.sin(angle * Math.PI / 180);
			break;
		}			
		case Robot.SPIN_LEFT:
		{
			double angle = Vector2D.get_angle_360(Vector2D.X_AXIS,
					robot.orientation);
			angle -= 5 + (Math.random() - 0.5) * 2*ANGLE_NOISE;
			robot.orientation.x = Math.cos(angle * Math.PI / 180);
			robot.orientation.y = Math.sin(angle * Math.PI / 180);
			
			double velocity = (Math.random() - 0.5) * 2;
			robot.position.x = robot.position.x + velocity
					* robot.orientation.x;
			robot.position.y = robot.position.y + velocity
					* robot.orientation.y;
			
			break;
		}
		case Robot.SPIN_RIGHT:
		{
			double angle = Vector2D.get_angle_360(Vector2D.X_AXIS,
					robot.orientation);
			angle += 5 + (Math.random() - 0.5) * 2*ANGLE_NOISE;
			robot.orientation.x = Math.cos(angle * Math.PI / 180);
			robot.orientation.y = Math.sin(angle * Math.PI / 180);

			double velocity = (Math.random() - 0.5) * 2;
			robot.position.x = robot.position.x + velocity
					* robot.orientation.x;
			robot.position.y = robot.position.y + velocity
					* robot.orientation.y;

			break;
		}
		}
	}

	
	
		
	static public void resolve_collisions(){
		List<Entity> entities = world.entities();
		for(int i=0; i<entities.size()-1; i++){
			Entity entity0 = entities.get(i); 
			for(int j=i+1; j<entities.size(); j++){
				Entity entity1 = entities.get(j);
				
				Vector2D vec = new Vector2D(entity1.position, entity0.position);
				double distance = vec.length();
				if (distance  >	entity1.r + entity0.r)
					continue;
				
				double ratio = 0.5;
				if (ui.MainPanel.picked_entity == entity0)
					ratio = 0;
				else if (ui.MainPanel.picked_entity == entity1)
					ratio = 1;
				vec = Vector2D.multiply(vec, (entity1.r+entity0.r-distance)/distance);
				entity0.position.add(Vector2D.multiply(vec, ratio));
				entity1.position.subtract(Vector2D.multiply(vec, 1-ratio));
			}
		}
	}
	
	
	
	
	// ͂̕ւƂ
	public static void resolve_collisions_static_forces(){
		List constraints = new ArrayList();
		
		List<Entity> entities = world.entities();
		int n = entities.size();
		
		Vector2D[] forces = new Vector2D[n];
		//Vector2D[] angle_forces = new Vector2D[n];
		double[] angle_forces = new double[n];
		for(int i=0; i<n; i++){
			forces[i] = new Vector2D();
			//angle_forces[i] = new Vector2D();
		}
		for(int i=0; i<entities.size()-1; i++){
			Entity entity0 = entities.get(i); 
			entity0.orientation.normalize();
		}
		
		// Փ˂
		for(int i=0; i<entities.size()-1; i++){
			Entity entity0 = entities.get(i); 
			for(int j=i+1; j<entities.size(); j++){
				Entity entity1 = entities.get(j);
				
				Vector2D vec = new Vector2D(entity1.position, entity0.position);
				double distance = vec.length();
				if (distance  >	entity1.r + entity0.r)
					continue;

				// collision!
				double ratio = entity1.weight / (entity0.weight+entity1.weight);
				if (ui.MainPanel.picked_entity == entity0)
					ratio = 0;
				else if (ui.MainPanel.picked_entity == entity1)
					ratio = 1;
				vec = Vector2D.multiply(vec, (entity1.r+entity0.r+1-distance)/distance);

				forces[i].add(Vector2D.multiply(vec, ratio));
				forces[j].subtract(Vector2D.multiply(vec, 1-ratio));

			}
			
			
			
		}
		// world bounds
		for(int i=0; i<entities.size(); i++){
			Entity entity0 = entities.get(i); 
			if (entity0.position.x < world.bbox.x + entity0.r)
				forces[i].add(new Vector2D(entity0.r - entity0.position.x, 0));
			if (entity0.position.x > world.bbox.x+world.bbox.width - entity0.r)
				forces[i].add(new Vector2D(world.bbox.x+world.bbox.width - entity0.r - entity0.position.x, 0));
			if (entity0.position.y < world.bbox.y + entity0.r)
				forces[i].add(new Vector2D(0, entity0.r - entity0.position.y));
			if (entity0.position.y > world.bbox.y+world.bbox.height - entity0.r)
				forces[i].add(new Vector2D(0, world.bbox.y+world.bbox.height - entity0.r - entity0.position.y));
		}
	
		

		for(int i=0; i<entities.size(); i++){
			Entity entity = (Entity) entities.get(i);
			entity.position.x += forces[i].x*0.5;
			entity.position.y += forces[i].y*0.5;
			

			entity.orientation.rotate_radian(angle_forces[i]*0.5);
			entity.orientation.normalize();
		}
		
		
	}
	// compute angle between center->v0 and center->v1
	static double get_angle(Vector2D center, Vector2D v0, Vector2D v1){

		Vector2D vec0 = new Vector2D(center, v0);
		Vector2D vec1 = new Vector2D(center, v1);
		double cosine = Vector2D.cos(vec0, vec1);
		double sine = Vector2D.sin(vec0, vec1);
		
		if (cosine <= -1) 
			return Math.PI;
		else if (cosine >= 1) 
			return 0;
		else {
			if (sine>0)
				return Math.acos(cosine);
			else 
				return -Math.acos(cosine);
		}
		    
	}
	
	
	
	
	
	
	
	
	
	// ŏ@go[W
	// ]`ߎĂ̂łȂƂɂȂ@iςɑ傫ȉ]j
	//@ƁAڐG_̈vɂȂ̂ŁAȕֈB
	public static void resolve_collisions_constraint(){
		List constraints = new ArrayList();
		
		// variables = [dx, dy, dt, dx, dy, dt....]
		// Ax = b;
		int n = world.entities().size();
		
		List<Entity> entities = world.entities();
		

		
		// Փ˂
		for(int i=0; i<entities.size()-1; i++){
			Entity entity0 = entities.get(i); 
			for(int j=i+1; j<entities.size(); j++){
				Entity entity1 = entities.get(j);
				
				Vector2D vec = new Vector2D(entity1.position, entity0.position);
				double distance = vec.length();
				if (distance  >	entity1.r + entity0.r)
					continue;

				// collision!
				
				// vector from center of entity0 to contact point
				Vector2D d0 = Vector2D.multiply(Vector2D.normalize(Vector2D.subtract(entity1.position, entity0.position)),
												entity0.r+1);
				Vector2D d1 = Vector2D.multiply(Vector2D.normalize(Vector2D.subtract(entity0.position, entity1.position)),
						entity1.r+1);
				
				Vector2D p0 = entity0.position;
				Vector2D p1 = entity1.position;
				
				// constaraint 
				// p0+t0*d0+v0 - p1+t1*d1+v1 = 0 
				double[] cx = new double[n*3+1];	// W
				cx[i*3+0] = 1;						// v0.x
				cx[i*3+2] = d0.y;					// t0
				cx[j*3+0] = -1;						// v1.x
				cx[j*3+2] = - d1.y;					// t1
				cx[n*3] = p0.x + d0.x - p1.x - d1.x;	// constant
				
				double[] cy = new double[n*3+1];	// W
				cy[i*3+1] = 1;						// v0.y
				cy[i*3+2] = -d0.x;					// t0
				cy[j*3+1] = -1;						// v1.x
				cy[j*3+2] =  d1.x;					// t1
				cy[n*3] = p0.y + d0.y - p1.y - d1.y;	// constant

				constraints.add(cx);
				constraints.add(cy);
			}
		}

		if (constraints.size() == 0){
			//System.out.println("no collision");
			return;
		}
		//System.out.println();
		
		// ܂Ȃׂ[ɋ߂Â鐧
		double w = 0.1;
		for(int i=0; i<entities.size(); i++){
			if (entities.get(i) instanceof Robot)
				w = 1;
			else
				w = 0.1;
			// v0.x = 0;
			double[] cx = new double[n*3+1];	// W
			cx[i*3+0] = w;						// x	
			constraints.add(cx);				
			double[] cy = new double[n*3+1];	// W
			cy[i*3+1] = w;						// y	
			constraints.add(cy);				
			double[] ct = new double[n*3+1];	// W
			ct[i*3+2] = w;						// t	
			constraints.add(ct);				
		}
		
		// Ŝŏ@ŉ
		double[][] A = new double[constraints.size()][n*3];
		double[] b = new double[constraints.size()];
		for(int i=0; i<constraints.size(); i++){
			double[] c = (double[]) constraints.get(i);
			for(int j=0; j<n*3; j++){
				A[i][j] = c[j];
				//System.out.print(", "+((int)(c[j]*10))/10.0);
			}
			b[i] = -c[n*3];
			//System.out.println(" = "+((int)(b[i]*10))/10.0);
		}
		
		
		double[][] AT = DoubleMatrix.transpose(A);
		double[][] ATA = DoubleMatrix.multiply(AT, A);
		double[] ATb = DoubleMatrix.multiply(AT, b);
		
		try{
		double[] x = DoubleMatrix.solve(ATA, ATb);
		
		for(int i=0; i<entities.size(); i++){
			Entity entity = (Entity) entities.get(i);
			entity.position.x += x[i*3+0];
			entity.position.y += x[i*3+1];
			
			double t = x[i*3+2] * 0.5;
			double[][] R = {{1, t}, {-t, 1}};
			entity.orientation = Vector2D.multiply(R, entity.orientation);
			entity.orientation.normalize();
		}
		}catch(Exception e){e.printStackTrace();};
		
	}
}


